from typing import List, Type, Callable, Optional

from starlette.routing import Route

from .. import Container


class StarletteContainer(Container):
    _request_singletons: List[Type]

    def __init__(self, request_singletons: Optional[List[Type]] = None, container=None):
        self._request_singletons = request_singletons or []
        super().__init__(container)

    def route(
        self,
        path: str,
        endpoint: Callable,
        *,
        methods: List[str] = None,
        name: str = None,
        include_in_schema: bool = True,
    ) -> Route:
        wrapped_endpoint = self.partial(endpoint, shared=self._request_singletons)
        return Route(
            path,
            wrapped_endpoint,
            methods=methods,
            name=name,
            include_in_schema=include_in_schema,
        )



from dataclasses import dataclass

from fastapi.params import Depends

from lagom.integrations.fast_api import FastApiContainer


@dataclass
class ComplexDep:
    something: str


def test_the_fast_api_container_can_return_a_fastapi_dependency():
    fac = FastApiContainer()
    fac[ComplexDep] = ComplexDep("testing")
    dependency_injection = fac.depends(ComplexDep)
    assert isinstance(dependency_injection, Depends)
    assert dependency_injection.dependency() == ComplexDep("testing")



import functools
import inspect
from copy import copy
from typing import Dict, Type, Union, Any, TypeVar, Callable, Set, List, Optional

from .interfaces import SpecialDepDefinition
from .exceptions import UnresolvableType, DuplicateDefinition
from .definitions import normalise, Singleton, Construction
from .util.reflection import RETURN_ANNOTATION
from .wrapping import bound_function

UNRESOLVABLE_TYPES = [str, int, float, bool]

X = TypeVar("X")


class Container:
    _registered_types: Dict[Type, SpecialDepDefinition]
    _explicitly_registered_types: Set[Type]

    def __init__(self, container: Optional["Container"] = None):
        self._registered_types = {}
        self._explicitly_registered_types = set()
        if container:
            self._registered_types = copy(container._registered_types)

    def define(
        self,
        dep: Type[X],
        resolver: Union[
            Type[X], Callable[[], X], Callable[[Any], X], SpecialDepDefinition[X], X
        ],
    ) -> None:
        if dep in self._explicitly_registered_types:
            raise DuplicateDefinition()
        self._registered_types[dep] = normalise(resolver)
        self._explicitly_registered_types.add(dep)

    def resolve(self, dep_type: Type[X], suppress_error=False) -> X:
        try:
            if dep_type in UNRESOLVABLE_TYPES:
                raise UnresolvableType(dep_type)
            registered_type = self._registered_types.get(dep_type, dep_type)
            return self._build(registered_type)
        except UnresolvableType as inner_error:
            if not suppress_error:
                raise UnresolvableType(dep_type) from inner_error
            return None  # type: ignore

    def partial(
        self, func: Callable[..., X], keys_to_skip=None, shared: List[Type] = None
    ) -> Callable[..., X]:
        if shared:
            return self._partial_with_shared_singletons(func, shared)
        spec = inspect.getfullargspec(func)

        def _bind_func():
            bindable_deps = self._infer_dependencies(
                spec, suppress_error=True, keys_to_skip=keys_to_skip or []
            )
            return functools.partial(func, **bindable_deps)

        return bound_function(_bind_func, func)

    def clone(self):
        return Container(self)

    def __getitem__(self, dep: Type[X]) -> X:
        return self.resolve(dep)

    def __setitem__(
        self,
        dep: Type[X],
        resolver: Union[
            Type[X], Callable[[], X], Callable[[Any], X], SpecialDepDefinition[X], X
        ],
    ):
        self.define(dep, resolver)

    def _build(self, dep_type: Any) -> Any:
        if isinstance(dep_type, SpecialDepDefinition):
            return dep_type.get_instance(self._build, self)
        return self._reflection_build(dep_type)

    def _reflection_build(self, dep_type: Type[X]) -> X:
        spec = inspect.getfullargspec(dep_type.__init__)
        sub_deps = self._infer_dependencies(spec)
        try:
            return dep_type(**sub_deps)  # type: ignore
        except TypeError as type_error:
            raise UnresolvableType(dep_type) from type_error

    def _infer_dependencies(
        self, spec: inspect.FullArgSpec, suppress_error=False, keys_to_skip=None
    ):
        keys_to_skip = keys_to_skip or []
        sub_deps = {
            key: self.resolve(sub_dep_type, suppress_error=suppress_error)
            for (key, sub_dep_type) in spec.annotations.items()
            if key != RETURN_ANNOTATION
            and sub_dep_type != Any
            and key not in keys_to_skip
        }
        filtered_deps = {key: dep for (key, dep) in sub_deps.items() if dep is not None}
        return filtered_deps

    def _partial_with_shared_singletons(
        self, func: Callable[..., X], shared: List[Type]
    ):
        def _cloned_container():
            temp_container = self.clone()
            # For each of the shared dependencies resolve before invocation
            # and replace with a singleton
            for type_def in shared:
                temp_container[type_def] = Singleton(
                    Construction(lambda: self.resolve(type_def))
                )
            return temp_container

        def _bind_func():
            temp_container = _cloned_container()
            return temp_container.partial(func, shared=[])

        return bound_function(_bind_func, func)



from typing import NewType, List

try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol  # type: ignore

from lagom import Container

MyServiceUrl = NewType("MyServiceUrl", str)


class ApiClientOfSomeKind:
    def __init__(self, url: MyServiceUrl):
        pass


class LoadBalancePerhaps:
    def __init__(self, urls: List[MyServiceUrl]):
        pass


class BakerProtocol(Protocol):
    def bake(self) -> str:
        pass


class Bakery:
    def __init__(self, baker: BakerProtocol):
        pass


class CrumpetBaker:
    def bake(self) -> str:
        return "crumpets"


def test_newtype_works_well_with_the_container(container: Container):
    container[MyServiceUrl] = MyServiceUrl("https://roll.diceapi.com")
    api_client = container[ApiClientOfSomeKind]
    assert isinstance(api_client, ApiClientOfSomeKind)


def test_lists_of_newtype_work_well_with_the_container(container: Container):
    container[List[MyServiceUrl]] = [
        MyServiceUrl("https://roll.diceapi.com"),
        MyServiceUrl("https://lol.diceapi.com"),
    ]
    load_balancer = container[LoadBalancePerhaps]
    assert isinstance(load_balancer, LoadBalancePerhaps)


def test_protocols_work_well_with_the_container(container: Container):
    # TODO: update contracts so this type ignore can be removed
    container[BakerProtocol] = CrumpetBaker  # type: ignore
    bakery = container[Bakery]
    assert isinstance(bakery, Bakery)



import inspect
from typing import TypeVar, Callable

INTROSPECTION_ATTRS = (
    "__module__",
    "__name__",
    "__qualname__",
    "__doc__",
    "__annotations__",
)

T = TypeVar("T")


def decorated_wrapper(wrapped: T, original) -> T:
    for attr in INTROSPECTION_ATTRS:
        if hasattr(original, attr):
            setattr(wrapped, attr, getattr(original, attr))
    return wrapped


def bound_function(
    function_builder: Callable[[], Callable], original_function: Callable
):
    if inspect.iscoroutinefunction(original_function):

        async def _wrapped_function(*args, **kwargs):
            return await function_builder()(*args, **kwargs)

    else:

        def _wrapped_function(*args, **kwargs):
            return function_builder()(*args, **kwargs)

    return decorated_wrapper(_wrapped_function, original_function)



import typing

from fastapi import Depends

from .. import Container

T = typing.TypeVar("T")


class FastApiContainer(Container):
    def depends(self, dep_type: typing.Type[T]) -> T:
        return Depends(lambda: self.resolve(dep_type))



from typing import Type, List, Optional

from flask import Flask

from lagom import Container


class FlaskContainer(Container):
    flask_app: Flask
    _request_singletons: List[Type]

    def __init__(
        self,
        app: Flask,
        request_singletons: Optional[List[Type]] = None,
        container=None,
    ):
        self.flask_app = app
        self._request_singletons = request_singletons or []
        super().__init__(container)

    def route(self, rule, **options):
        def decorator(f):
            endpoint = options.pop("endpoint", None)
            injected_func = self.partial(f, shared=self._request_singletons)
            self.flask_app.add_url_rule(rule, endpoint, injected_func, **options)
            return f

        return decorator
